/*
 * Simple inetd server.
 *
 * sinetd [-h <host>] <port> <command> [<args>...]
 *
 * Example:
 * sinetd 10000 /bin/ls -l
 * sinetd -h 127.0.0.1 10000 /bin/sh
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_GETHOSTBYNAME
#include <netdb.h>
#else
#include <arpa/inet.h>
#endif

static int create_socket(char *hostname, int port)
{
  int s, ret, on = 1;
  struct sockaddr_in address;
#ifdef USE_GETHOSTBYNAME
  struct hostent *host = NULL;
#endif

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0) {
    perror("socket()");
    exit(-1);
  }

  ret = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
  if (ret < 0) {
    perror("setsockopt()");
    exit(-1);
  }

  memset(&address, 0, sizeof(address));
  address.sin_family = AF_INET;
  address.sin_port = htons(port);
  if (hostname) {
#ifdef USE_GETHOSTBYNAME
    host = gethostbyname(hostname);
    if (host == NULL)
      perror("gethostbyname()");
    memcpy(&address.sin_addr, host->h_addr, host->h_length);
#else
    address.sin_addr.s_addr = inet_addr(hostname);
#endif
  } else {
    address.sin_addr.s_addr = htonl(DEFAULT_HOSTADDR);
  }

  ret = bind(s, (struct sockaddr *)&address, sizeof(address));
  if (ret < 0) {
    perror("bind()");
    exit(-1);
  }

  ret = listen(s, 5);
  if (ret < 0) {
    perror("listen()");
    exit(-1);
  }

  return s;
}

static int writestr(int s, char *buf)
{
  write(s, buf, strlen(buf));
  return 0;
}

static void sigchld_handler(int val)
{
  int status;
  wait(&status);
}

static int server_main(int s, int argc, char *argv[], char *envp[])
{
  int sockt;
  int pid;

  signal(SIGCHLD, sigchld_handler);

  while (1) {
    sockt = accept(s, NULL, NULL);
    if (sockt < 0) {
      continue;
    }

    pid = fork();
    if (pid < 0) {
      writestr(s, "fail fork.\n");
      exit(1);
    }
    if (pid == 0) { /* child process */
      close(s);
      dup2(sockt, 0);
      dup2(sockt, 1);
#ifdef CONNECT_STDERR_TO_SOCKET
      dup2(sockt, 2);
#endif
      close(sockt);
      if (execve(argv[0], argv, envp) < 0) {
	writestr(1, "fail exec.\n");
	_exit(1);
      }
      _exit(0);
    }

    /* parent process */
    close(sockt);
  }

  return 0;
}

static int help()
{
  fprintf(stderr, "sinetd [-h <host>] <port> <command> [<args>...]\n");
  exit(1);
}

int main(int argc, char *argv[], char *envp[])
{
  int port, s;
  char *hostname = NULL;

  argc--; argv++;

  if ((argc > 1) && (!strcmp(*argv, "-h"))) {
    argc--; argv++;
    hostname = *argv; argc--; argv++;
  }

  if (argc < 2) {
    help();
  }

  port = atoi(*argv); argc--; argv++;

  s = create_socket(hostname, port);
  server_main(s, argc, argv, envp);

  return 0;
}
